/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 ***************************************************************************
 * MODULE: Errormsg.c
 * error messages and logging
 * log:
 *  --message on stderr and on stdout if "Echo_on_stdout" is set
 *  --if the severity is high enough, will also log using syslog().
 *  --saves the last "non-Debug" log message
 * logerr:
 *  --formats the error message corresponding to "errno" and calls log();
 * fatal:
 *  -- log() and exit() with Errorcode
 * logerr_die:
 *  -- logerr() and exit() with Errorcode
 **************************************************************************/

#include "lp.h"

/*
 * Errormsg( err ) returns a printable form of the errormessage
 * corresponding to errno.
 */

char *
Errormsg( err )
	int err;
{
	char *cp;
	static char msgbuf[30];

	if( err >= 0 && err <= sys_nerr ){
		cp = sys_errlist[err];
	} else {
		(void)sprintf(msgbuf,"errno=%d",err);
		cp = msgbuf;
	}
	return(cp);
}

struct msgkind {
	int *var;
	char *str;
};
static struct msgkind msg_name[] = {
	{ &XLOG_ERR, "LOG_ERR" },
	{ &XLOG_CRIT, "LOG_CRIT" },
	{ &XLOG_WARNING, "LOG_WARNING" },
	{ &XLOG_NOTICE, "LOG_NOTICE" },
	{ &XLOG_INFO, "LOG_INFO" },
	{ &XLOG_DEBUG, "LOG_DEBUG" },
	{ 0 }
};

static
char *
logmsg( kind )
	int kind;
{
	int i;
	static char b[35];

	for( i = 0; msg_name[i].var; ++i ){
		if( *msg_name[i].var == kind ){
			return( msg_name[i].str );
		}
	}
	(void)sprintf(b, "bad message type %d",  kind );
	return(b);
}

/*
 * log( int severity; char *msg; args )
 *  	1. format the message; msg and args are similar to printf
 *		2. output on stderr
 *		3. if "Echo_on_stdout" output on stdout
 *		4. if severity high enough,  log using syslog()
 *		5. if not a Debug message, save in Last_errormsg[]
 */
/*VARARGS2*/
log(kind, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
int kind;
char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	char buf[BUFSIZ];

	plp_block_signals();

	buf[0] = 0;
	if(Name && *Name){
		(void)sprintf(buf+strlen(buf),"%s: ", Name);
	}
	if(Printer && *Printer){
		(void)sprintf(buf+strlen(buf),"%s- ", Printer);
	}
	if(Debug){
		(void)sprintf(buf+strlen(buf),"pid=%d, %s, ",getpid(),logmsg(kind));
	}
	(void)sprintf(buf+strlen(buf), msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);

#ifdef APPEND_TIME_TO_SYSLOG
	(void)sprintf(buf+strlen(buf)," at %s",Time_str() );

	if (kind <= XLOG_INFO)
		use_syslog( kind, buf);
#else
	if (kind <= XLOG_INFO)
		use_syslog( kind, buf);

	(void)sprintf(buf+strlen(buf)," at %s",Time_str() );
#endif

	if(Echo_on_stdout){
		(void)fprintf(stdout,"%s\n", buf);
		(void)fflush(stdout);
	}
	(void)fprintf(stderr,"%s\n", buf);
	(void)fflush(stderr);
	if (kind != XLOG_DEBUG ){
		(void)strcpy( Last_errormsg, buf );
	}

	plp_unblock_signals();

}

/*VARARGS2*/
fatal(kind, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	log( kind, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	cleanup();
	exit(Errorcode);
}

/*VARARGS2*/
logerr( kind,msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	int err = errno;
	char buf1[BUFSIZ];

	(void)sprintf(buf1, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	log( kind,"%s - %s", buf1, Errormsg(err) );
	errno = err;
}

/*VARARGS2*/
logerr_die( kind,msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	logerr( kind,msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	cleanup();
	exit(Errorcode);
}

/*
 * use the syslog(8) deamon to do our logging
 */
#ifdef NOSYSLOG
static FILE *f;
#else /* NOSYSLOG */
#include <syslog.h>
#endif /* NOSYSLOG */
static int init;

use_syslog( kind, msg )
	int kind;
	char * msg;
{
	if(Debug>7){
		(void)fprintf(stderr, "syslog: %s\n", msg );
		(void)fflush(stderr);
	}
#ifdef NOSYSLOG
	if( f != NULL || (f = fopen( "/dev/console", "w" ) ) != NULL ){
		(void)fprintf( f, "syslog: LPD %s\n", msg );
		(void)fflush(f);
	}
#else /* NOSYSLOG */
	if(init == 0 ){
		init = 1;
	}
# ifdef IS_VAX
		(void)syslog(SYSLOG_FACILITY|kind, msg);
# else
		(void)syslog(LOG_PID|SYSLOG_FACILITY|kind, "%s", msg);
# endif
#endif /* NOSYSLOG */
}
