/*
 * setuid.c: routines to manipulate user-ids securely (and hopefully,
 * portably). The internals of this are very hairy, because (a) there's
 * lots of sanity checking (if -D is active, at least); and (b) there's
 * at least three different setuid-swapping semantics to support :(
 *
 * Note that the main function saves the errno, then restores it afterwards;
 * this means it's safe to do "root_to_user();some_syscall();user_to_root();"
 * and use the errno.
 * 
 * in the diagrams:
 *     "root" is the user who owns the setuid executable (privileged).
 *     "user" is the user who runs it.
 *     "daemon" is an arbitrary user who owns data files used internally
 *            by the PLP utilities (spool directories, etc).
 *
 * root_to_user();	-- set euid from root to user
 * user_to_root();	-- set euid from user to root
 * root_to_daemon();	-- set euid from root to daemon
 * daemon_to_root();	-- set euid from daemon to root
 * full_root_to_user()	-- set euid *and uid* to user -- lose all privileges
 *
 * - Justin Mason <jmason@iona.ie> May '94.
 */ 

#include "lp.h"

static uid_t root_uid = 0;	/* root's uid number */
static uid_t user_uid;		/* the real uid */
/* there's also Daemon_uid, uid for "daemon" user, set by Get_Daemon() */

#define I_AM_UNSET	0
#define I_AM_ROOT	1
#define I_AM_USER	2
#define I_AM_DAEMON	3

static char *i_am_str[] = { "(unset)", "root", "user", "daemon" };

static int am_i_root = I_AM_UNSET; /* a flag for the *_daemon functions */

/* yuck -- HP/UX cpp doesn't seem to support #elif. duh. */

#ifdef USE_SETEUID
# define change_euid(from,to) (seteuid (to))
#else
# ifdef USE_SETRESUID
#  define change_euid(from,to) (setresuid (-1, to, -1))
# else
#  ifdef USE_SETREUID
#   define change_euid(from,to) (setreuid (from, to))
#  else
#   define change_euid(from,to) (setuid (to))	/* won't work on SVR4, at least */
#  endif
# endif
#endif

/**************************************************************/

void seteuid_backend (from, to, new_state)
uid_t from, to; int new_state; {
  int tmp_errno;

  if (from == to) {
    am_i_root = new_state;
    return;
  }
  tmp_errno = errno;	/* save it */

  if (Debug>2) {
    log (XLOG_DEBUG, "set-euid: from (%d, \"%s\") to (%d, \"%s\")",
 	 	from, i_am_str[am_i_root], to, i_am_str[new_state]);

    if (from != geteuid()) {	/* do this sanity check if we're debugging */
      log (XLOG_DEBUG, "set-euid: from (%d, \"%s\") is not euid! (euid=%d)",
		from, i_am_str[am_i_root], geteuid());
    }
  }
  if (Debug>5) log (XLOG_DEBUG, "before euid change: (%d, %d)", getuid(), geteuid());

  if (from != 0) {
    if (change_euid(from,0) < 0) {
      logerr_die( XLOG_INFO, "set-euid failed (NFS -nosuid option?)");
    }
  }
  if (Debug>5) log (XLOG_DEBUG, "from->0 result: (%d, %d)", getuid(), geteuid());

  if (to != 0) {
    if (change_euid(0,to) < 0) {
      logerr_die( XLOG_INFO, "set-euid failed");
    }
  }
  if (Debug>5) log (XLOG_DEBUG, "0->to result: (%d, %d)", getuid(), geteuid());

  am_i_root = new_state;
  errno = tmp_errno;	/* restore it */
}

void root_to_user ()   { seteuid_backend(0, user_uid, I_AM_USER); }

void root_to_daemon () { seteuid_backend(0, Daemon_uid, I_AM_DAEMON); }

void user_to_root ()   { seteuid_backend(user_uid, 0, I_AM_ROOT); }

void user_to_daemon () { seteuid_backend(user_uid, Daemon_uid, I_AM_DAEMON); }

void daemon_to_root () { seteuid_backend(Daemon_uid, 0, I_AM_ROOT); }

void daemon_to_user () { seteuid_backend(Daemon_uid, user_uid, I_AM_USER); }

/*************************************************************
 * set both uid and euid to the same value, using setuid().
 */

void setuid_backend (to)
int to; {
  if (Debug>2) {
    log (XLOG_DEBUG, "set-uid: from (%d, \"%s\") to uid=%d, euid=%d (no setuid privs)",
 	 	0, i_am_str[am_i_root], to, to);

    if (geteuid() != 0) {	/* do this sanity check if we're debugging */
      log (XLOG_INFO, "set-uid: euid != root! (euid=%d)", geteuid());
    }
  }
  if (setuid(to) < 0) {
    logerr_die( XLOG_INFO, "setuid failed");
  }
  /* okay, uid and euid are both set now. We can exec safely */
}

void full_root_to_user ()   { setuid_backend(user_uid); }
void full_root_to_daemon () { setuid_backend(Daemon_uid); }

/*************************************************************/

void set_user_uid(is_this_lpd)
int is_this_lpd; {
  if (is_this_lpd) {
    user_uid = Daemon_uid;
  } else {
    user_uid = getuid();
  }
}

/*************************************************************
 * Lose the dodgy environment variables, in case we should
 * go on to exec() stuff.
 * This is secure on SunOS and Solaris (so far).
 */

#ifndef HAS_UNSETENV
void unsetenv();
#endif

char *Env_secure_path = SECURE_PATH_ENV;
char *Env_secure_sunos_lib_path = SECURE_LD_PATH_ENV;

void
lose_insecure_envs()
{
  char *new_path, *new_lib_path;

#ifdef USE_PUTENV
  /* these two don't get deleted until the proc exits. see putenv(3). */

  new_path = (char *) malloc(strlen(Env_secure_path) + 1);
  new_lib_path = (char *) malloc(strlen(Env_secure_sunos_lib_path) + 1);

  if (putenv(new_path))		logerr_die("lose_insecure_envs: putenv");
  if (putenv(new_lib_path))	logerr_die("lose_insecure_envs: putenv");
#endif

  unsetenv("IFS");
  unsetenv("LD_PRELOAD");
  unsetenv("LD_PROFILE");
}

/* unsetenv -- from the sendmail 8 source. remove a variable
 * from the environment. */

#ifndef HAS_UNSETENV
void
unsetenv(name)
        char *name;
{
        extern char **environ;
        register char **pp;
        int len = strlen(name);

        for (pp = environ; *pp != NULL; pp++)
        {
                if (strncmp(name, *pp, len) == 0 &&
                    ((*pp)[len] == '=' || (*pp)[len] == '\0'))
                        break;
        }

        for (; *pp != NULL; pp++)
                *pp = pp[1];
}
#endif

/*************************************************************
 * a secure wrapper for the system() function.
 * edits out *ALL* shell metacharacters to avoid
 * embedded commands getting executed.
 */

static char *strndup(str, len)		/* strndup - make copy of string */
char *str; int len;
{
    char   *dst;

    if ((dst = malloc(len + 1)) == 0)
        logerr ("secure_system: malloc() failed");
    dst[len] = 0;
    return ((char *) strncpy(dst, str, len));
}

#ifdef _POSIX_ARG_MAX
#define MAX_ARGS	_POSIX_ARG_MAX
#else
#define MAX_ARGS	4096
#endif

static char metach[] = "$[]<>*?&^|`;{}'\"\\";
static char blnks[] = " \t\r\n";
static char meta_or_blnks[] = " \t\r\n$[]<>*?&^|`;{}'\"\\";

void ss_fork_and_or_exec (args, fork_flag)
char **args;
int fork_flag;	/* are we supposed to fork here? */
{
#if defined(INT_STATUS)
  int stat;
#else
  union wait stat;			/* daughter status */
#endif
  pid_t i, pid;

  if (fork_flag && (pid = fork())) {
    if (pid < 0) {
      logerr_die("secure_system: fork failed");
    }
    
    while ((i = wait(&stat)) > 0 && i != pid) {
      if(Debug>3)log(XLOG_DEBUG, "secure_system: caught %d, (%s)", i,
	      Decode_status(&stat));
    }
    if(Debug>3)log( XLOG_DEBUG, "secure_system: %d finished (%s)",
	      i,Decode_status(&stat) );
    return;
  }

  execvp(args[0], args);
  logerr_die("secure_system: execvp failed");
}

int secure_system_backend(cmd, fork_flag)
  char *cmd;
  int fork_flag;
{
  char *arg_p, *start_arg, *args[MAX_ARGS];
  int i, got_nonwhite, no_args, len;

  no_args = 0; got_nonwhite = 0;
  start_arg = cmd; arg_p = cmd;

  while (*arg_p) {
    if (no_args >= MAX_ARGS - 1) {
      log (XLOG_INFO,
		"secure_system: too many args (>= %d): \"%s\"", MAX_ARGS, cmd);
    }

    arg_p += strspn(arg_p, blnks);	/* skip whitespace */
    if (!*arg_p)
      break;

    /* if this matches, it's a metacharacter. Make big noise about
     * this possible hack attempt and return without running the command.
     */
    if (strchr(metach, *arg_p)) {
      log (XLOG_CRIT, "secure_system: attempted shell metachars: \"%s\"", cmd);
      return(1);
    }

    len = strcspn(arg_p, meta_or_blnks);
    args[no_args++] = strndup(arg_p, len);
    arg_p += len;
  }
  args[no_args] = 0;

#if 0
  for (i=0; args[i]; i++) {
    printf ("arg %d: \"%s\"\n", i, args[i]);
  }
#endif

  if (args[0]) {
    ss_fork_and_or_exec(args, fork_flag);
  } else {
    log (XLOG_INFO, "secure_system: no arguments!");
  }
  return(0);
}

int secure_system(cmd) { secure_system_backend(cmd, 1); }
int secure_exec(cmd)   { secure_system_backend(cmd, 0); }
