/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: lpd.c
 * lpd main program
 ***************************************************************************
 * lpd -- line Printer daemon.
 * 1. Get options 
 * 2. Check for existence of another daemon and exit if it exists.
 *    Note: this is done by locking a file
 * 3. Initialize error logging
 * 4. Start up all the queue handlers
 * 5. Set up communication sockets
 * 6. Loop forever doing:
 *     Listen for a connection and perform the requested operation.
 *     Operations are:
 *	    \1Printer\n
 *	    	check the queue for jobs and print any found.
 *	    \2Printer\n
 *	    	receive a job from another machine and queue it.
 *	    \3Printer [users ...] [jobs ...]\n
 *	    	return the current state of the queue (short form).
 *	    \4Printer [users ...] [jobs ...]\n
 *	    	return the current state of the queue (long form).
 *	    \5Printer Person [users ...] [jobs ...]\n
 *	    	remove jobs from the queue.
 *	    \6Printer Person operation
 *	    	enable/disable queueing, etc.
 *    Note: a process is forked for each operation
 *
 * Strategy to maintain protected spooling area:
 *	1. Spooling area is writable only by daemon and daemon group
 *	2. lpr runs setuid ROOT; it uses the user UID
 *	   to access any file it wants (verifying things before
 *	   with access(2)) and sets ownership of files in the spooling
 *     directory to be owner DAEMON, group DAEMON.
 *	3. Critical files in spooling area are owned by DAEMON, group DAEMON
 *	   with mode 660.
 *	4. lpd, lpq and lprm run setuid ROOT, and change group to DAEMON
 *	   Users can't get to anything w/o help of lpq and lprm programs.
 *	5. LPD forks a server process to service each device
 *	   in the printcap entry.  These processes run setuid root
 *	   and setgrp to their PID.  They open neccessary files, and
 *	   fork "filter" processes to process the output.  Filter
 *	   processes run UID DAEMON.
 *	6. The server process group is set to the main server PID.  This is used
 *	   by the LPRM program, which signals them with killpg(2)
 *
 * Error reporting:
 *	lpd will open a standard logging file. If not already present,
 *      no messages will be logged.  Critical messages will also be logged
 *      using syslog(8).
 *	The individual daemon processes will open log files also.  If
 *	they cannot be opened, then no logging will be done for the process.
 ***************************************************************************/

#include "lp.h"
#include "patchlevel.h"
#include <syslog.h>

SIG_RETURN_TYPE	segv_oops();		/* called when coredumping... makes noise */
SIG_RETURN_TYPE	cleanup();		/* file system scullery maid, cleans up files */
static pid_t lpdpid;			/* lpd process id */
static int finet;			/* socket for incoming requests */

int Inetd_mode;				/* invoked by inetd? */
int Queue_mode;				/* just handle queue management? */
int Single_printer;			/* debug the queue management? */
int Foreground;				/* standalone: no queue mgmt or forking */

static void Startup();

#if (SOCK_FAMILY==AF_UNIX)		/* networking addresses */
	struct sockaddr_un sa_us, sa_them;
#else
	struct sockaddr_in sa_us, sa_them;
#endif

void standalone_main () {
	int f;				/* Acme Integer, Inc. strikes again */
	int defreadfds, readfds;	/* select()ed connections */
	int nfds;			/* accept return value */
	int req;			/* request on this socket */
	int port_num;			/* used to get the inet server information */
	int fromlen;
#ifdef SO_REUSEADDR
	int one = 1;
#endif

	/*
	 * check to see that the is not a spooler present.  The LO is used
	 * when we exit to clean out the lock file information.
	 * Note that we check first, as the initial process, and then we fork
	 * another process.  This allows the output message to be printed
	 * on the stdout device in a synchronous manner.
	 */
	LO = Masterlock;
	if( (Lfd = Getlockfile(LO, &f, (char *)0, 0, (struct stat *)0)) == NULL ){
		if (Inetd_mode) {
			logerr_die (XLOG_INFO, "lpd: active LPD in standalone mode, pid %d", f);
			exit(1);
		} else {
			(void)fprintf( stdout, "active LPD, pid %d\n", f );
			exit(0);
		}
	}
	(void)fclose(Lfd);

	if (!Foreground) {
		/* We can fork and let the parent die */
		if ((f = fork()) < 0) {
			logerr_die( XLOG_INFO, "lpd: fork failed" );
		} else if( f ){
			exit(0);
		}

		Setuplog(Lpdlogf, 0 );	/* We can now start logging... */
	}

	if ((Lfd = Getlockfile(LO, &f, (char *)0, 0, (struct stat *)0)) == NULL) {
		logerr_die (stdout, "active LPD, pid %d\n", f);
		exit(1);
	}			/* ...and locking */

	/* save the PID of the LPD process;  it will be useful later
	 */
	lpdpid = getpid();

	if (!Foreground) {
		/* set pgrp to pid, so we can kill our kids later; also, 
		 * drop the control tty.  This prevents the user which started
		 * LPD from inadvertenly generation a signal and clobbering LPD.
		 */
#ifdef USE_SETSID
		if (setsid() == -1)
			logerr_die (XLOG_INFO, "setsid failed");
#else
#ifdef TIOCNOTTY
		if( (f = open("/dev/tty", O_RDWR, 0) ) >= 0) {
			(void)ioctl(f, TIOCNOTTY, (struct sgttyb *)0);
			(void)close(f);
		}
#endif
		if (setpgrp(lpdpid, 0) == -1)
			logerr_die (XLOG_INFO, "setpgrp failed");
#endif /* USE_SETSID */
	}

	/*
	 * set up signals; note that SIGPIPE may not be needed, but it
	 * never hurts to make sure.
	 */
	(void)signal(SIGCHLD, Reapchild);
	(void)signal(SIGPIPE, SIG_IGN);
	(void)signal(SIGHUP, cleanup);
	(void)signal(SIGINT, cleanup);
	(void)signal(SIGQUIT, cleanup);
	(void)signal(SIGTERM, cleanup);
	(void)signal(SIGSEGV, segv_oops);

	/*
	 * Set up the IPC.  This is very ugly,  and is stolen directly from
	 * the original LPD code and the 4.3 BSD Interprocess Communication
	 * Tutorial, which seems to have borrowed it from RLOGIN, etc.
	 *
	 * Set up INET socket
	 */
	if( ( finet = socket(SOCK_FAMILY, SOCK_STREAM, 0) ) < 0 ){
		logerr_die( XLOG_CRIT,"lpd: cannot create socket");
	}
	port_num = Link_port_num();

#ifdef SO_REUSEADDR
	/* set SO_REUSEADDR if it's supported */
	if ((setsockopt(finet, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof(one))) == -1) {
		logerr_die(XLOG_CRIT,"lpd: could not set socket option");
	}
#endif

	/* zero the sockaddr structure before using it */
	bzero( (char *)&sa_us, sizeof(sa_us) );

#if (SOCK_FAMILY==AF_UNIX)
	sa_us.sun_family = SOCK_FAMILY;
	sprintf(sa_us.sun_path, "/tmp/lpdsock%d", port_num);

	f = bind(finet, (struct sockaddr *)&sa_us,
			sizeof(sa_us.sun_family)+strlen(sa_us.sun_path);

	if (f < 0) {
		logerr_die( XLOG_CRIT,"lpd: bind failed domain: %d %s",
			      sa_us.sun_family, sa_us.sun_path);
	} 
#else
#if (SOCK_FAMILY==AF_INET)
	sa_us.sin_family = SOCK_FAMILY;
	/* the INADDR_ANY will allow different nets to be used */
	sa_us.sin_addr.s_addr = INADDR_ANY;
	sa_us.sin_port = port_num;

	f = bind(finet, (struct sockaddr *)&sa_us, sizeof(sa_us));

	if (f < 0) {
		logerr_die( XLOG_CRIT,"lpd: bind to internet socket failed");
	} 
#else
	sa_us.sin_family = SOCK_FAMILY;
	sa_us.sin_addr.s_addr = INADDR_ANY;
	sa_us.sin_port = port_num;

	f = bind(finet, (struct sockaddr *)&sa_us, sizeof(sa_us));

	if (f < 0) {
		logerr_die( XLOG_CRIT,"lpd: bind failed in unknown domain");
	} 
#endif
#endif

	set_user_uid(1); root_to_daemon ();	/* adopt safe perms (main) */

	/* put PID and time started in the lockfile */
	Setlockfile(LO, Lfd, lpdpid,Time_str());

	if( listen(finet, 5) < 0 ){
		logerr_die( XLOG_CRIT, "lpd: listen failed internet domain" );
	}
	if(Debug>2)log( XLOG_DEBUG, "lpd: socket %d", finet );

	/* Set up the sockets on which we will do an accept */
	defreadfds = (1 << finet);

	Startup(); /* Restart all the Printers. */

	/* Main loop: accept, do a request, continue.  */
	for (;;) {
		/*
		 * Defensive Coding a la Brain Damaged Unix Implementations
		 * A particular UNIX implementation will not cause SIGCHLD
		 * signals in a particularly reliable way.  Apparently if there
		 * is a SIGCHLD while it is masked off (error printing does this)
		 * IT IS IGNORED!!!  So we put a Reapchild() here;  at least
		 * we will gather them up each time we start a server
		 */
		(SIG_RETURN_TYPE) Reapchild();
		if(Debug>2)log( XLOG_DEBUG,"lpd: starting select");

		readfds = defreadfds;
		nfds = select(20, (fd_set *)&readfds, (fd_set *)0, (fd_set *)0,
			(struct timeval *)0);		 /* wait patiently */
		if (nfds <= 0) {
			if (nfds < 0 && errno != EINTR) {
				logerr_die( XLOG_CRIT,"lpd: select error");
			}
			continue;
		}
		/*
		 * incredible, but true: could have NO requests
		 */
		if( readfds ){
			if(Debug>2)log( XLOG_DEBUG,"lpd: doing select 0x%x", readfds );
			/* get the socket causing the select */
			fromlen = sizeof(sa_them);
			req = accept(finet, (struct sockaddr *)&sa_them, &fromlen);
			if (req < 0) {
				logerr( XLOG_CRIT,"lpd: finet accept");
				continue;
			}
			if (!From_host()){
				logerr(XLOG_INFO,"lpd: From_host failed");
				continue;
			}
			/*
			 * Fork process to handle activity requested
			 */
			(void) fflush(stdout);
			(void) fflush(stderr);
			if ((f = fork()) == 0) {	/* daughter */
				(void) fclose(Lfd);
				(void) close(finet);
				Lfd = NULL;
				if( dup2(req, 1) < 0 ){
					logerr_die( XLOG_CRIT, "lpd: dup2 for server failed" );
				}
				(void) close(req);
				servicereq();
				(void) close(1);
				exit(0);
			} else if( f < 0 ){
				logerr( XLOG_CRIT, "lpd: fork failed");
			}
			(void) close(req);
		}
	}
}

/*
 * inetd_main: handle one request and exit.
 */

#if (SOCK_FAMILY==AF_INET)
void inetd_main () {
	int f;
	int len;			/* length of the sockaddr structures */
	int port_num;			/* used to get the inet server information */

	Setuplog(Lpdlogf, 1);		/* start logging, but don't close stdin */

	set_user_uid(1); root_to_daemon ();	/* adopt safe perms (main) */
	port_num = Link_port_num();

	/* save the PID of the LPD process;  it will be useful later */
	lpdpid = getpid();

	/*
	 * set up signals; note that SIGPIPE may not be needed, but it
	 * never hurts to make sure.
	 */
	(void)signal(SIGCHLD, Reapchild);
	(void)signal(SIGPIPE, SIG_IGN);
	(void)signal(SIGHUP, cleanup);
	(void)signal(SIGINT, cleanup);
	(void)signal(SIGQUIT, cleanup);
	(void)signal(SIGTERM, cleanup);
	(void)signal(SIGSEGV, segv_oops);

	len = sizeof(sa_them);
	if (getpeername (0, (struct sockaddr *) &sa_them, &len) == -1) {
		logerr_die(XLOG_INFO, "lpd: getpeername failed");
	}
	if (!From_host()) {		/* failure due to insecure remote port */
		logerr(XLOG_INFO, "lpd: From_host failed");
		exit(0);
	}

	len = sizeof(sa_us);
	if(getsockname(0, (struct sockaddr *) &sa_us, &len) == -1) {
		logerr_die(XLOG_INFO, "lpd: getsockname failed");
	}

	if( dup2(0, 1) < 0 ){
		logerr_die( XLOG_CRIT, "lpd: dup2 for server failed" );
	}
	servicereq();	/* this'll read from fd 1 */
	close(0);
	close(1);
}
#endif

/*
 * queue-managing mode: just fork off the queue lpds.
 */
void queue_main () {
	int f;				/* Acme Integer, Inc. strikes again */

	/*
	 * check to see that the is not a spooler present.  The LO is used
	 * when we exit to clean out the lock file information.
	 */
	set_user_uid(1); root_to_daemon ();	/* adopt safe perms (main) */
	LO = Masterlock;
	if ((Lfd = Getlockfile(LO, &f, (char *)0, 0, (struct stat *)0)) == NULL) {
		(void)fprintf( stdout, "active LPD, pid %d\n", f );
		exit(1);
	}

	if (Single_printer) {
		Startprinter();
	} else {
		Startup();
	}
	Closelockfile(LO, Lfd);
	Lfd = 0;
	exit(0);
}

main(argc, argv)
	int argc;
	char **argv;
{
	(void)umask(0);		/* explicitly set umask */
	Getoptions(argc, argv);	/* process command line options */
	Std_environ();
#	ifdef XPERIMENT
	Setup_test();
#	endif /* XPERIMENT */
	Tailor_names ();	/* set up the pathnames for information files */

	if (Inetd_mode) {
		inetd_main();
	} else if (Queue_mode) {
		queue_main();
	} else {
		standalone_main();	/* never exits */
	}
	exit(0);
}

/***************************************************************************
 * Getoptions(argv, argc)
 *    char **argv; int argc;
 * Purpose:
 * extracts options and arguments from command line using Getopt(2)
 * Side Effects:
 *	-X : calls setup_test() to modify defaults 
 *  -D nn : sets Debug level
 *  -L file : sets log file
 ***************************************************************************/

static char usage_msg[] =
	"usage: lpd [-iqv] [-D dbg] [-L log] [-P printer] [-F]\n(\"lpd -v\" for more help)\n";

static void show_version()
{
	extern char *Env_secure_path, *Env_secure_sunos_lib_path;

	printf ("PLP version %s\n\n", PLP_PATCHLEVEL);

	printf("by Patrick Powell, U. Minnesota (up to version 2.0)\n");
	printf("and Justin Mason, Iona Technologies <jmason@iona.ie> (from 3.3 on)\n\n");

	printf("%s\narguments:\n", usage_msg);
#if (SOCK_FAMILY==AF_INET)
	printf("        -i inetd mode\n");
#endif
	printf("        -q queue-processing mode\n");
	printf("        -v version info and extra usage (this message)\n");
	printf("        -D debug level (if compiled with -DDEBUG)\n");
	printf("        -L alternate log file (default is %shostname)\n", Lpdlogf);
	printf("        -P printer (handle only one queue)\n");
	printf("        -F operate in foreground\n");

	printf("Secure executable search path for filters and pre-filters:\n");
	printf("        \"%s\"\n", Env_secure_path);
	printf("Secure shared-libraries search path for filters and pre-filters:\n");
	printf("        \"%s\"\n\n", Env_secure_sunos_lib_path);

	printf("Enforce file-size limitations? (USE_MX): %s\n",
#ifdef USE_MX
				"yes");
#else
				"no");
#endif

	printf("Prohibit printing of \"non-text\" files? (ALLOW_TEXT_ONLY): %s\n\n",
#ifdef ALLOW_TEXT_ONLY
				"yes");
#else
				"no");
#endif

	exit(0);
}

Getoptions(argc, argv)
	int argc;
	char **argv;
{
	int c;
	Inetd_mode = 0;
	Queue_mode = 0;
	Single_printer = 0;
	Foreground = 0;
	while ((c = Getopt(argc, argv, "XiqvD:L:P:F")) != EOF){
		switch(c){
			case 'X':	/* test version */
				Setup_test();
				Tailor_names ();
				break;
			case 'D':		/* turn on Debugging */
				if( Optarg == NULL ){
		    			(void)fprintf(stderr, usage_msg);
					exit(1);
				}
				Debug= atoi( Optarg );
				break;
			case 'L':
				if( Optarg == NULL ){
		    			(void)fprintf(stderr, usage_msg);
					exit(1);
				}
				(void)strcpy(Lpdlogf, Optarg);
				break;
#if (SOCK_FAMILY==AF_INET)
			case 'i': /* inetd mode */
				Inetd_mode = 1;
				break;
#endif
			case 'q': /* queue management mode */
				Queue_mode = 1;
				break;
			case 'P': /* queue debugging mode */
				Single_printer = 1;
				Queue_mode = 1;
				if( Optarg == NULL ) {
		    			(void)fprintf(stderr, usage_msg);
					exit(1);
				}
    				Printer = malloc(strlen(Optarg)+1);
				(void)strcpy(Printer, Optarg);
				break;
			case 'F': 
				Foreground = 1;
				break;
			case 'v':
				show_version();
				break;
			default:
		    		(void)fprintf(stderr, usage_msg);
				exit(1);
		}
	}
	if( Optind < argc ){
		(void)fprintf( stderr, "%s: extra argument %s\n", Name, argv[Optind] );
		exit(1);
	}
}
/***************************************************************************
 * Setuplog( char *logfile, int saveme )
 * Purpose: to set up a standard error logging environment
 * saveme will prevent stdin from being clobbered
 *   1.  open /dev/null on fd 0 unless saveme is set
 *   2.  dup 0 to fd 1 unless saveme is set
 *   3.  If logfile is "-" or NULL, output file is alread opened
 *   4.  Open logfile; if unable to, then open /dev/null for output
 ***************************************************************************/
Setuplog( logfile, saveme )
	int saveme;
	char *logfile;
{
	int fd;

	/*
	 * we want 0 (stdin) to be /dev/null
	 */
	(void)fflush(stdout);
	(void)fflush(stderr);
	if( saveme != 1) {
		(void)close(0);
		if( (fd = open( "/dev/null", O_RDWR, 0 )) != 0 ){
			logerr_die( XLOG_CRIT, "Setuplog: /dev/null opened as %d", fd);
		}
		/*
		 * do the dup
		 */
		(void)close(1);
		if( dup2(0,1) < 0){
			logerr_die( XLOG_CRIT, "Setuplog: dup2 failed" );
		}
	}
	if(Debug>4)log(XLOG_DEBUG,"Setuplog: opening log file %s", logfile );
	/*
	 * open logfile;  if it is "-", use stderr
	 */
	if( logfile && *logfile && strcmp(logfile, "-")) {
		if( (fd = open (logfile, O_WRONLY|O_APPEND, 0)) < 0 ){
			if(Debug>0)logerr(XLOG_DEBUG,"cannot open logfile %s",logfile);
			/* write to stderr if debugging, /dev/null otherwise */
			if(Debug>0){
				fd = 2;
			} else {
				fd = 0;
			}
		}
		if( fd != 2 && dup2(fd,2) < 0){
			logerr_die( XLOG_CRIT, "Setuplog: dup2 failed" );
		}
		if( fd > 2 ){
			(void)close(fd);
		}
	}
}

/*
 *	When a child dies, it will generate a SIGCHLD,
 *	and Reapchild() will lay the body to rest with a wait3().
 */
SIG_RETURN_TYPE Reapchild()
{
	int pid;
#if defined(INT_STATUS)
	int status;
#else
	union wait status;
#endif

	while (
#ifdef USE_WAITPID
			(pid = waitpid((pid_t) -1, &status, WNOHANG))
#else
			(pid = wait3(&status, WNOHANG, (struct rusage *)0))
#endif
				 > 0) {

		if(Debug>3)log(XLOG_DEBUG,"process %d, status %s", pid,
			Decode_status( &status ) );
	}
}

/*
 *	Clean up; kill off children and remove sockets and files
 */
SIG_RETURN_TYPE cleanup()
{
	int pid;

	plp_block_signals();

	if( Lfd ){
		Closelockfile(LO, Lfd);
		Lfd = 0;
	}
	pid = getpid();
	if(Debug>0)log( XLOG_DEBUG, "cleanup" );
	/*
	 * if the recfiles server, clean up files
	 */
	rm_recfiles();

	if( pid == lpdpid) {			/* we are session leader */
		log(XLOG_CRIT, "cleanup: lpd process %d terminating", pid);
#ifdef USE_KILLPG
    killpg(lpdpid,SIGKILL);			/* honey, I killed the kids */
#else
    kill(-lpdpid,SIGKILL);
#endif
		shutdown (finet, 2);		/* emergency shutdown socket */
		close (finet);			/* close it */
	}

	/*
	 * Wait for children until the bodies rot
	 */
	(SIG_RETURN_TYPE) Reapchild();
	exit(Errorcode);
}

/***********************************************************************
 * From_host()
 *   This routine will extract the remote Host Name and remote Host
 *   port number information.  The remote process must connect on
 *   a port in a "reserved" range;  these are reserved for use
 *   by priviledged processes (setuid/uid root).
 * Side Effects: sets the "from" string to the Name of the remote Host.
 * Returns: 1 if privileged and successful, 0 if not.
 * NOTE: this code was derived from the orignal 4.2BSD LPR source,
 *   the 4.3BSD Interprocess Communications Tutorials, and other places.
 ***********************************************************************/

#if (SOCK_FAMILY==AF_UNIX)
From_host()
{
	if (From != NULL) {
		if (Debug>4) log(XLOG_INFO, "From_host: freeing old From");
		free(From);
	}
	if( (From = malloc( (unsigned)strlen(Host)+1 )) == 0 ){
		logerr_die(XLOG_CRIT,"malloc failed in From_host" );
	}
	(void)strcpy(From, Host);
	if(Debug>4)logerr(XLOG_INFO,"From_host: Host Name: '%s'", From);
	return(1);
}
#else
From_host()
{
        struct hostent *hp;                     /* Host entry */
        extern char *inet_ntoa();       /* inet address translation routine */
        int port;                                       /* from port number */

        port = ntohs(sa_them.sin_port);
        if (sa_them.sin_family != SOCK_FAMILY) {
                logerr_die(XLOG_CRIT,
			"From_host: malformed from address, family %d, not right %d",
			sa_them.sin_family, SOCK_FAMILY);
        }

	hp = gethostbyaddr((char *)(&sa_them.sin_addr),
			sizeof(struct in_addr), sa_them.sin_family);
	if (hp == 0) {
		char *s;	/* temp string */
		s = inet_ntoa(sa_them.sin_addr);
		logerr(XLOG_INFO,"From_host: Hostname for address '%s' unknown",s);

		if ((From = (char *) malloc( (unsigned)strlen(s)+1)) == 0) {
			logerr_die(XLOG_CRIT,"malloc failed in From_host" );
		}
		(void)strcpy(From, s);
	} else {
		if ((From = (char *) malloc( (unsigned)strlen(hp->h_name)+1 )) == 0 ){
			logerr_die(XLOG_CRIT,"malloc failed in From_host" );
		}
		(void)strcpy(From, hp->h_name);
	}

#ifdef CHECK_INCOMING_PORT
        if (port > Maxportno) {
#ifdef STRICT_INCOMING_PORT
		log(XLOG_CRIT,"refused insecure request from host %s", From);
		return(0);
#else
		log(XLOG_INFO,"processing insecure request from host %s", From);
#endif
	}
#endif
	return(1);
}
#endif

static int
Printer_needs_start(printer)
char *printer;
{
  Printer = printer;
  if( Get_pc_entry(Printer, All_pc_vars, All_pc_len ) == 0){
    log( XLOG_INFO, "trying to start non-existent Printer" );
    return;
  }
  if( SD == 0 || *SD == 0 ){
	  /*
	   * no spooling directory, not a Printer
	   */
	  if(Debug>0)log(XLOG_DEBUG, "Startprinter: not a Printer");
	  exit(0);
  }
  /* Startprinter will do the other sanity checks */
  return(Getq());
}

/************************************************************************
 * Startup()
 * start queue servers
 ************************************************************************/
static void
Startup()
{
	int pid;
	char **pr;
#if defined(INT_STATUS)
	int status;
#else
	union wait status;
#endif

	/*
	 * Restart the server.
	 * Start scanning the printcap for entries
	 */
	if ((pid = fork()) < 0) {
		logerr( XLOG_CRIT,"queue server: cannot fork");
	} else if(pid == 0 ) {
		if(Debug>4)log( XLOG_DEBUG, "starting servers" );
		for( pr = All_printers(); *pr; ++pr ){
			if (!Printer_needs_start(*pr)) {
				if(Debug>2)log( XLOG_DEBUG,
					"%s server: empty queue, no need to fork", *pr);
				continue;
			}
			if ((pid = fork()) < 0) {
				logerr( XLOG_CRIT,"server for %s: cannot fork", Printer);
				continue;
			} else if(pid == 0) {
				if(Debug>2)log( XLOG_DEBUG, "started %s server, pid %d",
					Printer,getpid());
				(void)fclose(Lfd);
				Lfd = 0;
				Startprinter();
				exit(0);
			}
			sleep( (unsigned)1 );
		}
		while ((pid=wait(&status)) > 0){
			if(Debug>3)log(XLOG_DEBUG,"process %d, status %s", pid,
				Decode_status( &status ) );
		}
		if(Debug>4)log( XLOG_DEBUG, "all servers done" );
		exit(0);
	}
}

SIG_RETURN_TYPE segv_oops()
{
	int pid;

	plp_block_signals();

	if( Lfd ){
		Closelockfile(LO, Lfd);
		Lfd = 0;
	}
	pid = getpid();
	if(Debug>0)log( XLOG_DEBUG, "segv cleanup" );
	/*
	 * if the recfiles server, clean up files
	 */
	rm_recfiles();

	if( pid == lpdpid) {			/* we are session leader */
		log(XLOG_CRIT, "cleanup: lpd process %d seg faulted", pid);
#ifdef USE_KILLPG
    killpg(lpdpid,SIGKILL);			/* honey, I killed the kids */
#else
    kill(-lpdpid,SIGKILL);
#endif
		shutdown (finet, 2);		/* emergency shutdown socket */
		close (finet);			/* close it */
	}

	exit(Errorcode);
}
