/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: Print_support.c
 * handle the actual output to the Printer
 ***************************************************************************
 * void Print_open(): opens the Printer
 * static int Print_of_fd(): makes an 'of' filter if needed
 * void Print_close(): closes the Printer
 * int Print_ready(): combines Print_open() and Print_of_fd()
 * static int of_stop(): stops the 'of' filter
 * int Print_string( str ): prints a string through 'of' or to Printer
 * int Print_copy( fd ): copies a file through 'of' or to Printer
 * int Print_filter( file, cmd ): makes a filter and copies file
 * int Print_banner(): prints a banner through 'of' or to Printer
 * Note: the above procedures which return values return JSUCC on success,
 *   and JFAIL or JABORT on failure
 ***************************************************************************/

#include "lp.h"

/***************************************************************************
 * Print_open()
 * 	Open the Printer, and set the Print_fd variables
 * 	If the RW printcap flag is set, output is opened RW, otherwise
 * 	opened writeonly in append mode.
 * 
 * 	If the Printer is a tty (i.e.- isatty returns non-zero),
 * 	then the baud rate is set.  The baud rate table is stolen from
 * 	the original LPD code, which stole it from STTY, etc.
 * Side Effect:
 * 	sets the Print_fd variable to a non-zero value
 *  terminates if Printer is unable to be opened
 ****************************************************************************/

static int of_pid;		/* OF process pid */
static int of_fd;		/* pipe to OF process */
static int of_running;	/* of filter running */
static char filter_stop[] = "\031\001";	/* what to send filter */

static char tcp_host[HOSTLEN];	/* (*tcp_host != 0) if LP runs over a TCP socket */
static unsigned short int tcp_port;

void parse_LP_for_tcp();

Print_open()
{
	int err;

	/*
	 * check to see if it is open
	 */
	if( Print_fd ){
		return;
	}
	if( LP == 0 || *LP == 0 ){
		fatal( XLOG_INFO, "Missing LP value for local Printer");
	}
	setstatus("waiting for %s to become ready since %s (offline?)",
		Printer, Time_str());

	parse_LP_for_tcp();

	if (*tcp_host) {
		Print_fd = open_lp_sock();
	} else {
		daemon_to_root();	/* setuid to open the LP device */
		Print_fd = open(LP, RW ? ( O_RDWR|O_APPEND ) : ( O_WRONLY|O_APPEND ), 0);
		root_to_daemon();	/* setuid to open the LP device */
	}

	if( Print_fd < 0) {
		err = errno;
		setstatus("Print_open: '%s' open(%s) failed, %s",Printer,LP,
			Errormsg(err));
		errno = err;
		logerr_die( XLOG_INFO, "Print_open: cannot open %s", LP);
	} else if( Print_fd == 0 ){
		fatal( XLOG_INFO, "Print_open: cannot happen - Print_fd == 0!");
	}
	/*
	 * if it is a tty, set the baud rates and control information
	 */
	if (isatty(Print_fd) && !*tcp_host) {
		Do_stty(Print_fd);
	}
	if(Debug>3)log(XLOG_DEBUG,"Print_open: %s is fd %d", LP, Print_fd );
}

/*
 * Start up an output filter, if needed.
 */
Print_of_fd()
{
	int p[2];
	char *cmd;

	if( OF == 0 || *OF == 0 || of_pid ){
		return;
	}
	/*
	 * set OF command line
	 */
	cmd = Setup_filter('o',OF);
	if( cmd == 0 || *cmd == 0 ){
		fatal(XLOG_INFO, "bad OF entry");
	}
	if(Debug>2)log( XLOG_DEBUG, "starting OF, '%s'", cmd);
	/*
	 * create pipe and fork
	 */
	if( pipe(p) < 0 ){
		logerr_die( XLOG_NOTICE, "Print_of_of: pipe failed" );
	}
	if( p[0] < 3 || p[1] < 3 ){
		fatal( XLOG_INFO, "Print_of_fd: IMPOSSIBLE- pipe fd %d %d", p[0], p[1]);
	}
	if ((of_pid = fork()) == 0) {	/* child */
		if( dup2(p[0], 0) < 0		/* pipe is std in */
		 || dup2(Print_fd, 1) < 0 ){		/* Printer is std out */
			logerr_die( XLOG_NOTICE, "Print_of_fd: dup2 failed" );
		}
		if(Debug>3)log( XLOG_DEBUG, "Print_of_fd: forked %d OF: %s",
			getpid(), cmd );

		daemon_to_root(); full_root_to_daemon();
		secure_exec( cmd );		/* run as "daemon" */
		logerr_die( XLOG_NOTICE,"Print_of_fd: execv failed %s", cmd);
	} else if( of_pid < 0 ){
		logerr_die( XLOG_NOTICE, "Print_of_fd: fork failed" );
	}
	(void)close(p[0]);		/* close input side */
	of_fd = p[1];
	of_running = of_pid;
	if(Debug>3)log(XLOG_DEBUG,"started OF process %d, fd %d", of_pid, of_fd);
}

/***************************************************************************
 * Print_close()
 * 1. Close the pipes and outputs.
 * 2. Signal the of process that it is to do something.
 *	Note that there is a Reapchild() call that will collect the
 *	children;  this will make sure that there is a minimal number of
 *	outstanding processes active.
 ***************************************************************************/

void
Print_close()
{
	if (of_fd) {
		(void)close(of_fd);
		of_fd = 0;
	}
	if (Print_fd) {
		(void)close(Print_fd);	/* close Printer */
		Print_fd = 0;
	}
	/*
	 * when you kill the printer, make sure that you do it neatly
	 */
	if( of_pid ){
		(void)kill(of_pid, SIGCONT);
	}
	of_pid = 0;
}

/***************************************************************************
 * Print_ready()
 * 1. open the Printer if neccessary
 * 2. start up the output filter
 * 3. if the filter has been stopped, start it
 * Return: JSUCC if started, JFAIL if not
 ***************************************************************************/
Print_ready()
{
	/*
	 * open the Printer
	 */
	Print_open();
	Print_of_fd();
	if( of_pid && of_running == 0 ){
		if (kill(of_pid, SIGCONT) < 0) {
			logerr( XLOG_INFO,"cannot restart output filter");
			Print_close();
			return( JFAIL );
		}
		of_running = of_pid;
	}
	return( JSUCC );
}

/***************************************************************************
 * of_stop()
 * stop the output filter if neccessary
 * 1. open and create the filter
 * 2. flush output
 * 3. kill(SIGSTOP) it
 * Return: JSUCC if stopped, JFAIL if not
 ***************************************************************************/

int
of_stop()
{
#if defined(INT_STATUS)
	int statb;
#else
	union wait statb;
#endif
	int pid;

	Print_open();
	Print_of_fd();
	/*
	 * stop the OF filter
	 */
	if( of_pid && of_running ){
		if(Debug>3)log( XLOG_DEBUG, "stopping output filter" );
		if( write( of_fd, filter_stop, strlen(filter_stop))
			!= strlen(filter_stop) ){
			logerr( XLOG_NOTICE,"of_stop: cannot write to OF");
			Print_close();
			return( JFAIL );
		}
		/*
		 * wait until OF blocks
		 */
		if(Debug>3)log( XLOG_DEBUG, "of_stop: waiting for OF %d", of_pid );

		while ( 
#ifdef USE_WAITPID
                        (pid = waitpid((pid_t) -1, &statb, WUNTRACED))
#else
			(pid = wait3(&statb,WUNTRACED,(struct rusage *)0))
				 != -1
#endif

			&& pid != of_pid) ;
#if defined(INT_STATUS)
		if( pid < 0 || !WIFSTOPPED(statb)){
#else
		if( pid < 0 || statb.w_stopval != WSTOPPED ){
#endif
			logerr( XLOG_INFO, "of_stop: OF %d died (%s)", of_pid,
				Decode_status( &statb ) );
			Print_close();
			return( JFAIL );
		}
		if(Debug>3)log( XLOG_DEBUG, "of_stop: output filter stopped" );
		of_running = 0;
	}
	return( JSUCC );
}

/***************************************************************************
 * Print_string( char *str )
 * print a string through a filter 
 * 1. Enable the line Printer
 * 2. get the filter or device fd;
 * 3. put out the string
 * 4. if unsuccessful, close the Printer
 * Return: JSUCC if successful, JFAIL otherwise
 ***************************************************************************/
int
Print_string( str )
	char *str;
{
	int f;
	int l;

	if( Print_ready() != JSUCC ){
		return( JFAIL );
	}
	if( of_fd ){
		f = of_fd;
	} else {
		f = Print_fd;
	}
	l = strlen( str );
	if( write( f, str, l ) != l ){
		logerr( XLOG_INFO, "Print_string: write error");
		Print_close();
		return( JFAIL );
	}
	return( JSUCC );
}


/***************************************************************************
 * Print_copy( int fd )
 * copy a file through a filter
 * 1. ready the Printer
 * 2. copy the file to the appropriate output device
 * 3. if an error, close the Printer
 ***************************************************************************/
int
Print_copy( fd )
	int fd;
{
	int f;
	long cnt;
	char buf[BUFSIZ];
	int in;			/* bytes read */
	int out;		/* bytes written out */

	cnt = 0;
	if( Print_ready() != JSUCC ){
		return( JFAIL );
	}
	if( of_fd ){
		f = of_fd;
	} else {
		f = Print_fd;
	}
	while( (in = read( fd, buf, sizeof(buf) )) > 0 ){
		out = write( f, buf, in );
		if( in != out ){
			logerr( XLOG_INFO, "Print_copy: write error");
			Print_close();
			return( JFAIL );
		}
		cnt = cnt + out;
	}
	/*
	 * completed the reading
	 */
	if( in < 0 ){
		logerr( XLOG_INFO, "Print_copy: read error");
		Print_close();
		return( JFAIL );
	}
	if(Debug>3)log(XLOG_DEBUG,"Print_copy: printed %d bytes", cnt);
	return( JSUCC );
}

/***************************************************************************
 *Print_filter( int fd, char *cmd, int lflag)
 *  spawn a subprocess to do the printing
 * 1. stop the Printer
 * 2. fork a process
 * 3. wait for process to complete
 * 4. restart the process
 * Return: JSUCC if successful, JFAIL otherwise
 ***************************************************************************/
int
Print_filter( file, cmd, lflag )
	int file;
	char *cmd;
	int lflag;	/* is this run on a local printer? */
{
	int succ;			/* success code */
	int filter_pid;		/* filter process */
	int pid;			/* daughter pid */
#if defined(INT_STATUS)
	int status;
#else
	union wait status;	/* daughter status */
#endif
	int err;			/* saved value of errno */
	
	if (lflag) {
		/*
		 * stop the Printer
		 */
		succ = of_stop();
		if( succ != JSUCC ){
			return( succ );
		}
	}
	/*
	 * fork a process,  and connect file to fd 0, Printer to fd 1.
	 */
	if ((filter_pid = fork()) == 0) {	/* daughter */
		/*
		 * dup input file to standard in and Printer to stdout
		 */
		if( dup2(file, 0)<0 || dup2(Print_fd, 1)<0 ){
			logerr_die(XLOG_NOTICE,"Print_filter: dup2 failed filter %s",cmd);
		}

                daemon_to_root(); full_root_to_daemon();
		secure_exec(cmd);
		logerr_die( XLOG_NOTICE,"Print_filter: cannot execv %s", cmd);
	} else if( filter_pid < 0 ){
		logerr( XLOG_NOTICE, "Print_filter: fork failed" );
		if (lflag) {
			Print_close();
		}
		return( JFAIL );
	}
	/*
	 *	Wait for filter to complete 
	 */
	if(Debug>2)log(XLOG_DEBUG,"Print_filter: waiting for pid %d",filter_pid);
	while ((pid = wait(&status)) > 0 && pid != filter_pid){
		if(Debug>3)log( XLOG_DEBUG, "Print_filter:caught %d (%s)",
			pid, Decode_status(&status) );
	}
	err = errno;
	if(Debug>2)log( XLOG_DEBUG, "Print_filter: filter %d finished (%s)",
		pid,Decode_status(&status));
	errno = err;
	/*
	 * Check to see how filter terminated
	 */
	if( pid < 0 || !WIFEXITED(status)
		|| (unsigned)WEXITSTATUS(status) > 1 ){
		/*
		 * died for bad reasons, don't run this again
		 */
		log(XLOG_INFO,"Print_filter:Filter '%s' Malfunction (%s)",
			cmd, Decode_status(&status));
		if (lflag) {
			Print_close();
		}
		return( JABORT );
	} else if (WEXITSTATUS(status) != 0){
		/*
		 * try again
		 */
		log( XLOG_INFO, "Print_filter:Filter '%s' Retry wanted", cmd );
		if (lflag) {
			Print_close();
		}
		return(JFAIL);
	}
	return(JSUCC);
}

/*
 * Print_banner()
 * 1. get the Printer ready
 * 2. call the banner() routine with the correct parameter
 */
Print_banner()
{
	int f;

	if(Debug>3)log(XLOG_DEBUG,"Print_banner: printing banner");
	if( Print_ready() != JSUCC ){
		return( JFAIL );
	}
	if( of_fd ){
		f = of_fd;
	} else {
		f = Print_fd;
	}
	if( banner(f) != JSUCC ){
		if(Debug>3)log(XLOG_DEBUG,"Print_banner: banner failed");
		Print_close();
		return( JFAIL );
	}
	return( JSUCC );
}

void parse_LP_for_tcp()
{
  /* %hu -> unsigned short int */
  if (sscanf(LP, "%hu@%s", &tcp_port, tcp_host) == 2) {
    if(Debug>3)log(XLOG_DEBUG,"check_for_tcp: LP is a TCP socket");

  } else {
    if(Debug>3)log(XLOG_DEBUG,"check_for_tcp: LP is a local device");
    *tcp_host = '\0';
    tcp_port = 0;
  }
}

int open_lp_sock()
{
  int tmperr;
  int sockfd = -1;
  struct hostent *host;                   /* host entry pointer */
  struct sockaddr_in sin;

  tmperr = 0;

  if ((host = gethostbyname(tcp_host)) == NULL) {
	  logerr_die (XLOG_INFO,"open_lp_sock: unknown host %s", tcp_host);
  }

  bzero( (char *)&sin, sizeof( sin ));
  sin.sin_family = AF_INET;
  bcopy(host->h_addr, (caddr_t)&sin.sin_addr, host->h_length);
  sin.sin_family = host->h_addrtype;
  sin.sin_port = htons(tcp_port);

  do {
    if(Debug>3)log(XLOG_DEBUG,"open_lp_sock: host %s port %d", tcp_host, tcp_port);
    /*
     * Open a TCP socket.
     */
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      logerr_die(XLOG_INFO, "can't open inet socket\n");
    }
    if (connect(sockfd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
	setstatus("waiting for %s to become ready - offline? (%s)", Printer,
	    Errormsg(errno));
	close(sockfd);
	sleep(15);

    } else {
        if(Debug>3)log(XLOG_DEBUG,"got connection to %s, fd %d", LP, sockfd);
	break;
    }

  } while (1);		/* block forever */

  errno = tmperr;
  return sockfd;
}
