/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: Startprinter.c
 * Printer queue job handler
 * 
 * StartPrinter is responsible for handling the Printer queue.
 * 1.  Get the printcap entry.
 * 2.  Check the spool directory to see if printing is enabled
 * 3.  Check to see if there are any jobs
 * 4.  If a remote Printer, call remotePrinter()
 *     If a local Printer, call localPrinter()
 ***************************************************************************/

#include "lp.h"
extern int Printjob(), Sendjob();
extern int Printinit(), Setup_xfer();
extern void Printfinal(), Printerror(), Allsent(), Send_error();

extern int Single_printer;
extern void chdir_SD();

Startprinter()
{
	int pid;				/* process id */
	char *ps, *st;			/* ACME Pointers */

	/*
	 * ignore SIGCHILD,  explicitly wait for them
	 */
	(void)signal(SIGCHLD, SIG_DFL);
	/*
	 * get the printcap entry
	 */
	if( Get_pc_entry(Printer, All_pc_vars, All_pc_len ) == 0){
		log( XLOG_INFO, "trying to start non-existent Printer" );
		exit(0);
	}
	if( SS && *SS ){
		/*
		 * we have a spool server that is to be started
		 */
		ps = PS;
		st = ST;
		if( Set_pc_entry(SS, All_pc_vars, All_pc_len ) == 0){
			log( XLOG_INFO, "queue %s does not entry", SS );
			exit(0);
		}
		PS = ps;
		ST = st;
		LO = Printer;
		SV = 0;
	}
	if( SD == 0 || *SD == 0 ){
		/*
		 * no spooling directory, not a Printer
		 */
		if(Debug>0)log(XLOG_DEBUG, "Startprinter: not a Printer");
		exit(0);
	}
	if( RM && NW ){
		(void)Remote_start();
		exit(0);
	}
	chdir_SD();
	/*
	 * check to see if there is a daemon running
	 */
	Lfd = Getlockfile( LO, &pid, (char *)0, 0, &LO_statb );
	if( Lfd == NULL ){
		if(Debug>2)log(XLOG_DEBUG, "Startprinter: server present, pid %u",pid);
		exit(0);
	}
	if( fseek( Lfd, 0L, 0 ) < 0 ){
		logerr_die( XLOG_INFO, "Startprinter: T1 fseek failed");
	}
	/*
	 * check for unspooling enabled
	 */
	if( LO_statb.st_mode & DISABLE_PRINT) {
		if(Debug>0)log( XLOG_DEBUG,"Startprinter: unspooling disabled");
		exit(0);
	}
	/*
	 * handle multiple servers
	 */
	if( SV && *SV ){
		Closelockfile(LO, Lfd );
		Lfd = 0;
		Start_server();
		/*
		 * Check on the status of the queue lock file
		 */
		Lfd = Getlockfile( LO, &pid, (char *)0, 0, &LO_statb );
		if( Lfd == NULL ){
			if(Debug>2)log(XLOG_DEBUG, "Startprinter: server present, pid %u",pid);
			exit(0);
		}
	}
	/*
	 * set up log file
	 */
	if( fseek( Lfd, 0L, 0 ) < 0 ){
		logerr_die( XLOG_INFO, "Startprinter: T2 fseek failed");
	}
	if(Debug>0)log(XLOG_DEBUG, "Startprinter: should log into %s",
		LF?LF:"/dev/null");
        if (!Single_printer) {
		Setuplog( LF, 0 );
	}
	if( fseek( Lfd, 0L, 0 ) < 0 ){
		logerr_die( XLOG_INFO, "Startprinter: T3 fseek failed");
	}
        if (!Single_printer) {
	/*
	 * OK, you are the server, go to it.
	 * set process group to the queue handler process
	 */
#ifdef USE_SETSID
		if (setsid() == -1)
			logerr_die (XLOG_INFO, "setsid failed");
#else
		if (setpgrp(getpid(), 0) == -1)
			logerr_die (XLOG_INFO, "setpgrp failed");
#endif
	}

	Setlockfile( LO, Lfd, getpid(), "(checking queue)");
	setstatus( "checking queue at %s", Time_str());

	/*
	 * search the spool directory for work and sort by queue order.
	 */
	if(Debug>0)log(XLOG_DEBUG, "Startprinter: unqueueing jobs");
	if( RM ){
		unspool(Setup_xfer, Sendjob, Allsent, Send_error);
	} else {
		unspool(Printinit, Printjob, Printfinal,Printerror);
	}
	if(Debug>2)log( XLOG_DEBUG, "Startprinter: work done" );
	setstatus( "ready at %s", Time_str() );
	Closelockfile(LO,Lfd);
	exit( 0 );
}


/***********************************************************
 * unspool( init, jobhandler, closer)
 * 
 * do{
 *   check queue for new jobs
 *   jobdone = 0
 *   for(n = 0;n < Jobcount; ++n)
 *      if queue needs to be resorted then
 *         set jobdone and break from "for" loop;
 *      while not successful and retry count is less than a limit do
 *         init();
 *         success = jobhandler(job n);
 *     -- note: job can be busy (no action), done successfully,
 *              done unsuccessfully and should be abandoned,
 *              or done unsuccessfully and should be repeated
 *      if action was done then set jobdone and remove job
 *          else call closer();
 * }while( jobdone )
 * call closer()
 ***********************************************************/

unspool( init, dojob, closer, err )
	int (*init)();
	int (*dojob)();
	void (*closer)();
	void (*err)();
{
	int n;				/* job number */
	int retries;		/* number of retries */
	int success;		/* job success status */
	struct plp_queue *q;	/* job entry  */
	int jobdone;		/* job done in this pass */
	FILE *cfp;			/* control file */
	int init_done;	/* initialization done */

	/*
	 * the big loop
	 */
	init_done = 0;
	do{
		/*
		 * search the spool directory for more work.
		 */
		Jobcount = Getq();
		if(Debug>4)log(XLOG_DEBUG,"unspool: %d jobs", Jobcount);
		/*
		 * reset "update queue order" flag
		 */
		if(fstat(fileno(Lfd), &LO_statb ) < 0 ){
			logerr_die( XLOG_NOTICE,"unspool: cannot stat %s", LO);
		}
		if(Debug>5)log(XLOG_DEBUG,"unspool: '%s' perms 0%o",LO,
			LO_statb.st_mode);
		if(fchmod(fileno(Lfd), (int)(LO_statb.st_mode & CLEAR_REQUE)) < 0){
			logerr_die( XLOG_NOTICE,"unspool: cannot chmod '%s'", LO);
		}
		jobdone = 0;
		/*
		 * scan the queue
		 */
		if(Debug>4)log(XLOG_DEBUG,"unspool: running through jobs");
		for(n=0; n < Jobcount; ++n ){
			q = &Queue[n];
			/*
			 * lock the control file
			 */
			if( (cfp = Lockcf(q->q_name) )!= NULL ){
				if(Debug>4)log(XLOG_DEBUG,"unspool: %s starting", q->q_name);
				Setlockfile( LO, Lfd, getpid(), q->q_name );
				/*
				 * try printing this file until successful,
				 *  number of attempts exceeds limit, or abort
				 * try sending indefinately until successful
				 */
				success = JFAIL;
				for( retries = 0;
					success == JFAIL && retries <= RT;
					++retries ){
					/*
					 * STDIO file error reset, needed on some implementations
					 */
					clearerr( cfp );
					/*
					 * seek to the start of the file
					 */
					if( fseek( cfp, 0L, 0 ) < 0 ){
						logerr_die(XLOG_INFO,"unspool: fseek failed (%s)",
							q->q_name );
					}
					/*
					 * update status information
					 */
					if(Debug)log(XLOG_DEBUG,"unspool: doing %s, %d retry",
						q->q_name,retries);
					/*
					 * initialize
					 */
					if( init_done == 0 ){
						init_done = (*init)();
						if( init_done != JSUCC ){
							setstatus( "initialization failure, %s",Last_errormsg);
							return;
						}
					}
					setstatus( "printing, active job started %s, attempt %d",
						Time_str(), retries+1);
					/*
					 * try to print the job
					 */
					success = (*dojob)(cfp, q);
				}
				if(Debug>4)log(XLOG_DEBUG,"unspool: %s status %d", q->q_name,
					success );
				if( success == JFAIL ){
					(*err)();
					init_done = 0;
				}
				jobdone = 1;
				Remove_job( cfp, q );
				(void)fclose(cfp);
			} else {
				if(Debug>4)log(XLOG_DEBUG,"unspool: %s locked", q->q_name);
			}
			/*
			 * collect any garbage
			 */
			Reapchild();
		}
		/*
		 * collect any garbage
		 */
		Reapchild();
	} while( jobdone );
	/*
	 * collect any garbage
	 */
	Reapchild();
	/*
	 * queue finished
	 */
	if(Debug>0)log(XLOG_DEBUG, "unspool: no more work to do" );
	setstatus("ready at %s", Time_str());
	(*closer)();
}

/***************************************************************************
 * Start_server()
 * start multiple servers for a single queue.  The SV printcap parameter
 * has a list of the servers.  These servers use the spool queues
 * directory,  and have a lock file in the server directory.
 * This routine will extract the names of each server,  fork a process
 * for it,  and then will read the printcap information for the server
 * from the queue.
 * The printer names are separated by commas.
 ***************************************************************************/

Start_server()
{
	char *ps;
	char *last;
	int pid;
	
	last = 0;
	ps = SV;
	while( ps ){
		Printer = ps;
		if( last = index( ps, ',' ) ){
			*last = 0;
			ps = last+1;
		} else {
			ps = 0;
		}
		if( ps ){
			if( (pid = fork()) < 0 ){
				logerr_die( XLOG_INFO, "Start_server: fork failed" );
			} else if( pid == 0 ){
				ps = 0;
			}
		}
	}
	if( Set_pc_entry(Printer, All_pc_vars, All_pc_len ) == 0){
		fatal( XLOG_INFO, "trying to start non-existent server" );
	}
	LO = Printer;
	if(Debug>3)log(XLOG_DEBUG,"Start_server: started" );
}
