/***************************************************************************
 * U. Minnesota LPD Software * Copyright 1987, 1988, Patrick Powell
 * version 3.3.0 Justin Mason July 1994
 ***************************************************************************
 * MODULE: lpr_canprint.c
 * Checks to see if a file is printable
 *
 * This will be a system dependent function.  If you have setreuid(),
 * you can implement this in a fairly effective manner, as the file
 * open and stat() is done as the user, not as root.
 *
 * The "is_exec" and "is_arch" functions are very system dependent.
 * If you have the source for file(1) available,  check to see how it
 * determines the file types.  This may be more bizzare than you expect.
 ***************************************************************************/

#include "lpr.h"
extern long lseek();
static int is_arch(), is_exec();

/***************************************************************************
 * Is_printable(  char *f )
 * Test to see if this is a printable file.
 * Return: 0 if unprintable; 1 otherwise
 ***************************************************************************/
static char errmsg[] = "File '%s' not printed: %s";

int
Is_printable( file, statb )
	char *file;
	struct stat *statb;	/* file status */
{
	char buf[BUFSIZ];
	int fd = -1;		/* file descriptor */
	int n, i, c;		/* Acme Integers, Inc. */
	int succ = 0;

	/* do all this as the user */
	/* 3.3.0: we're user by default here */

	if( access(file, F_OK | R_OK) < 0){
		Warnmsg(errmsg, file, "cannot access it");
	} else if(Remove && access(file, F_OK | R_OK | W_OK ) < 0){
		Warnmsg(errmsg, file, "cannot remove it");
 	} else if ((fd = open(file, O_RDONLY, 0)) < 0) {
		Warnmsg(errmsg, file, "cannot open it");
	} else if (fstat(fd, statb) < 0) {
		Warnmsg(errmsg, file, "cannot stat it");
	} else if ((statb->st_mode & S_IFMT) == S_IFDIR) {
		Warnmsg(errmsg, file, "it is a directory");
	} else if (statb->st_size == 0) {
		Warnmsg(errmsg, file, "it is an empty file");
	} else if( (n = read(fd, buf, sizeof(buf))) <= 0 ){
		Warnmsg(errmsg, file, "cannot read it");
	} else if( Binary || Format == 'l' ){
		succ = 1;
	} else if( !(index( "fpr", Format) || (XT && index( XT, Format)) )){
		/*
		 * We don't have to do the following checks, applicable to
		 * text files.
		 */
			succ = 1;

	} else if( is_exec(fd, buf, n) ){
		Warnmsg(errmsg, file, "executable program");
	} else if( is_arch(fd, buf, n)){
		Warnmsg(errmsg, file, "archive file");
	} else {
		succ = 1;
#ifdef ALLOW_TEXT_ONLY
		/* check for clean and lovely files, up to the first block
		 * jmason: this will choke on MMDF-separated mail
		 * digests, anything else with ^Ms, etc. etc.; forget it.
		 */
		for( i = 0; i < n; ++i ) {
			c = buf[i];
			if(!isascii(c) || !( c == '\b' || isprint(c) || isspace(c))){
				Warnmsg(errmsg, file, "it is a garbage file");
				succ = 0;
				break;
			}
		}
#endif
	}
	if( succ == 0 ){
		(void)close(fd);
		fd = -1;
	} else if( lseek( fd, 0L, 0 ) < 0 ){
		logerr( XLOG_INFO, "Is_printable: lseek failed %s", file );
		(void)close(fd);
		fd = -1;
	}
	return(fd);
}
/*
 * The is_exec and is_arch are system dependent functions
 * which check if a file is an executable or archive file,
 * based on the information in the header.
 */

#ifdef	USE_A_OUT_H
#include <a.out.h>
#endif /* USE_A_OUT_H */

#ifdef USE_EXECHDR_H
#include <sys/exechdr.h>
#endif /* USE_EXECHDR_H */

#ifdef IS_DATAGEN
#include <sgs.h>
#endif

/*
 * The N_BADMAG macro isn't present on some OS'es.
 *
 * If it isn't defined on yours, just edit your system's
 * "system/os.h" file and #undef both USE_A_OUT_H and
 * USE_EXECHDR_H. This function's just a luxury.
 *
 * for some reason, Solaris 2.3 (for one) doesn't support
 * the N_BADMAG macro, even though it has all the values. duh.
 * fill it in here, based on the std BSD-style format.
 *
 * if NMAGIC isn't defined, N_BADMAG isn't going to work,
 * so disable it and the file-testing code below.
 */
#ifndef N_BADMAG
#ifdef NMAGIC
#define N_BADMAG(x) \
	((x).a_magic!=OMAGIC && (x).a_magic!=NMAGIC && (x).a_magic!=ZMAGIC)

#else /* no NMAGIC */
#undef USE_A_OUT_H
#undef USE_EXECHDR_H

#endif /* NMAGIC */
#endif /* N_BADMAG */

static int
is_exec( fd, buf, n )
	int fd;
	char *buf;
	int n;
{
	int i = 0;

#ifdef IS_DATAGEN
	struct header header;
	if( lseek(fd, (long)0, 0 ) < 0 )
		logerr(XLOG_INFO,"is_exec: cannot lseek");
	if( read(fd, (char *)&header, sizeof(header) ) == sizeof(header))
		i = ISMAGIC(header.magic_number);
#endif /* IS_DATAGEN */

#if !defined(USE_A_OUT_H) && !defined(USE_EXECHDR_H)
	return(0);

#else
	if( n >= sizeof( struct exec )) {
		i = !(N_BADMAG( (*(struct exec *)buf) ));
	}
	return(i);
#endif /* USE_etc. */
}

#include <ar.h>

static int
is_arch( fd, buf, n )
	int fd;
	char *buf;
{
#ifdef ARMAG
	return( !strncmp( buf, ARMAG, strlen( ARMAG )) );
#else
	return(0);
#endif /* ARMAG */
}
